import math

from jax import numpy as jnp
from jax.scipy import linalg

from lqg.model import System, Dynamics, Actor
from lqg.kalman import KalmanFilter


class DampedSpringTrackingFilter(KalmanFilter):
    def __init__(self, process_noise=1., sigma=6., dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        A = jnp.array([[1., 1.], [-k, lmbd]])
        C = jnp.array([[1., 0.]])

        V = jnp.diag(jnp.array([1e-7, sigma_w]))
        W = jnp.diag(jnp.array([sigma]))
        super().__init__(A, C, V, W)


class DampedSpringModel(System):
    def __init__(self, process_noise=.2, sigma=1., c=0.1, motor_noise=0.5,
                 dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        prop_noise = 0.

        # dynamics model
        A = jnp.array([[1., 0., 1.], [0., 1., 0], [-k, 0., lmbd]])

        B = dt * jnp.array([[0.], [10.], [0.]])

        # observation model
        C = jnp.array([[1., 0., 0.], [0., 1., 0]])
        # C = jnp.array([[0., 0., 0.] * delay + [1., -1., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, motor_noise, sigma_w]))
        W = jnp.diag(jnp.array([sigma, prop_noise]))
        # W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., 0.]]))
        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V)

        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DampedSpringCostlessModel(DampedSpringModel):
    def __init__(self, process_noise=.2, sigma=1., motor_noise=0.5, dt=1. / 60.):
        super().__init__(process_noise=process_noise, sigma=sigma, c=0., motor_noise=motor_noise, dt=dt)


class DampedSpringSubjectiveModel(System):
    def __init__(self, process_noise=.2, sigma=1., c=0.1, motor_noise=0.5, subj_noise=.2, subj_k=.0007, subj_lmbd=0.675,
                 dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        prop_noise = 0.

        # dynamics model
        A = jnp.array([[1., 0., 1.], [0., 1., 0], [-k, 0., lmbd]])
        A_subj = jnp.array([[1., 0., 1.], [0., 1., 0], [-subj_k, 0., subj_lmbd]])

        B = dt * jnp.array([[0.], [10.], [0.]])

        # observation model
        C = jnp.array([[1., 0., 0.], [0., 1., 0]])
        # C = jnp.array([[0., 0., 0.] * delay + [1., -1., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, motor_noise, sigma_w]))
        V_subj = jnp.diag(jnp.array([1e-7, motor_noise, subj_noise]))
        W = jnp.diag(jnp.array([sigma, prop_noise]))
        # W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., 0.]]))
        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V)

        act = Actor(A=A_subj, B=B, C=C, V=V_subj, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DampedSpringSubjectiveVelocityModel(System):
    def __init__(self, process_noise=.2, sigma=1., c=0.1, motor_noise=0.5, subj_noise=.2, subj_vel_noise=1.,
                 dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        prop_noise = 0.

        # dynamics model
        A = jnp.array([[1., 0., 1.], [0., 1., 0], [-k, 0., lmbd]])
        A_subj = jnp.array([[1., 0., dt], [0., 1., 0], [0., 0., 1.]])

        B = dt * jnp.array([[0.], [10.], [0.]])

        # observation model
        C = jnp.array([[1., 0., 0.], [0., 1., 0]])
        # C = jnp.array([[0., 0., 0.] * delay + [1., -1., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, motor_noise, sigma_w]))
        V_subj = jnp.diag(jnp.array([subj_noise, motor_noise, subj_vel_noise]))
        W = jnp.diag(jnp.array([sigma, prop_noise]))
        # W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., 0.]]))
        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V)

        act = Actor(A=A_subj, B=B, C=C, V=V_subj, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DampedSpringDiffModel(System):
    def __init__(self, process_noise=.2, sigma=1., c=0.1, motor_noise=0.5,
                 dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        prop_noise = 0.

        # dynamics model
        A = jnp.array([[1., 0., 1.], [0., 1., 0], [-k, 0., lmbd]])

        B = dt * jnp.array([[0.], [10.], [0.]])

        # observation model
        C = jnp.array([[1., -1., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, motor_noise, sigma_w]))
        W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0.], [-1., 1., 0.], [0., 0., 0.]]))
        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V)

        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DampedSpringTwoDimFullModel(System):
    def __init__(self, process_noise=.2, sigma_h=1., sigma_v=1., motor_noise_h=0.5, motor_noise_v=0.5,
                 c_h=.1, c_v=.1, prop_noise_h=0.5, prop_noise_v=0.5,
                 dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        # dynamics model
        A = jnp.array([[1., 0., 0., 0., 1., 0.],
                       [0., 1., 0., 0., 0., 0.],
                       [0., 0., 1., 0., 0., 1.],
                       [0., 0., 0., 1., 0., 0.],
                       [-k, 0., 0., 0., lmbd, 0.],
                       [0., 0., -k, 0., 0., lmbd]])

        B = dt * jnp.array([[0., 0.], [10., 0.], [0., 0.], [0., 10.], [0., 0.], [0., 0.]])

        # observation model
        C = jnp.eye(4, 6)
        # C = jnp.array([[0., 0., 0.] * delay + [1., -1., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, motor_noise_h, 1e-7, motor_noise_v, sigma_w, sigma_w]))
        W = jnp.diag(jnp.array([sigma_h, prop_noise_h, sigma_v, prop_noise_v]))
        # W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0., 0., 0., 0.],
                                         [-1., 1., 0., 0., 0., 0.],
                                         [0., 0., 1., -1., 0., 0.],
                                         [0., 0., -1., 1., 0., 0.],
                                         [0., 0., 0., 0., 0., 0.],
                                         [0., 0., 0., 0., 0., 0.]]))
        R = jnp.diag(jnp.array([c_h, c_v]))

        dyn = Dynamics(A=A, B=B, C=C, V=V)

        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DampedSpringTwoDimModel(DampedSpringTwoDimFullModel):
    def __init__(self, process_noise=.2, sigma_h=1., sigma_v=1., motor_noise_h=0.5, motor_noise_v=0.5,
                 c_h=.1, c_v=.1,
                 dt=1. / 60.):
        super().__init__(process_noise=process_noise, sigma_h=sigma_h, sigma_v=sigma_v,
                         motor_noise_h=motor_noise_h, motor_noise_v=motor_noise_v,
                         c_h=c_h, c_v=c_v, prop_noise_h=0., prop_noise_v=0.,
                         dt=dt)


class DampedSpringTwoDimCostlessModel(DampedSpringTwoDimModel):
    def __init__(self, process_noise=.2, sigma_h=1., sigma_v=1., motor_noise_h=0.5, motor_noise_v=0.5,
                 dt=1. / 60.):
        super().__init__(process_noise=process_noise, sigma_h=sigma_h, sigma_v=sigma_v,
                         motor_noise_h=motor_noise_h, motor_noise_v=motor_noise_v,
                         c_h=0., c_v=0.,
                         dt=dt)


class DampedSpringTwoDimSubjectiveModel(System):
    def __init__(self, process_noise=.2, sigma_h=1., sigma_v=1., motor_noise_h=0.5, motor_noise_v=0.5,
                 c_h=.1, c_v=.1, subj_noise=2.,
                 dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        # dynamics model
        A = jnp.array([[1., 0., 0., 0., 1., 0.],
                       [0., 1., 0., 0., 0., 0.],
                       [0., 0., 1., 0., 0., 1.],
                       [0., 0., 0., 1., 0., 0.],
                       [-k, 0., 0., 0., lmbd, 0.],
                       [0., 0., -k, 0., 0., lmbd]])

        B = dt * jnp.array([[0., 0.], [10., 0.], [0., 0.], [0., 10.], [0., 0.], [0., 0.]])

        # observation model
        C = jnp.eye(4, 6)
        # C = jnp.array([[0., 0., 0.] * delay + [1., -1., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, motor_noise_h, 1e-7, motor_noise_v, sigma_w, sigma_w]))
        V_subj = jnp.diag(jnp.array([1e-7, motor_noise_h, 1e-7, motor_noise_v, subj_noise, subj_noise]))
        W = jnp.diag(jnp.array([sigma_h, 0., sigma_v, 0.]))
        # W = jnp.diag(jnp.array([sigma]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0., 0., 0., 0.],
                                         [-1., 1., 0., 0., 0., 0.],
                                         [0., 0., 1., -1., 0., 0.],
                                         [0., 0., -1., 1., 0., 0.],
                                         [0., 0., 0., 0., 0., 0.],
                                         [0., 0., 0., 0., 0., 0.]]))
        R = jnp.diag(jnp.array([c_h, c_v]))

        dyn = Dynamics(A=A, B=B, C=C, V=V)

        act = Actor(A=A, B=B, C=C, V=V_subj, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


class DampedSpringVelocityModel(System):
    def __init__(self, process_noise=.2, sigma=1., c=0.01, motor_noise=0.1, dt=1. / 60.):
        lmbd = 0.675
        k = 0.0007
        sigma_w = process_noise

        At = jnp.array([[1., 1], [-k, lmbd]])

        tau1 = 224 / 1000.
        tau2 = 13 / 1000.

        # compute system dynamics and cost matrices
        A0 = jnp.array([[0., 1.], [-1 / (tau1 * tau2), -(tau1 + tau2) / (tau1 * tau2)]])
        Ar = linalg.expm(dt * A0)

        A = linalg.block_diag(At, Ar)

        idx = list(range(A.shape[0]))
        swapdim = idx[::2] + idx[1::2]

        A = A[:, swapdim]
        A = A[swapdim, :]
        # A = linalg.block_diag(A[::2, ::2], A[1::2, 1::2])

        B0 = jnp.array([[0.], [1. / (tau1 * tau2)]])

        # Taylor series approximation of B matrix
        taylor = dt * jnp.array(
            [jnp.linalg.matrix_power(A0 * dt, k) / float(math.factorial(k + 1)) for k in range(50)])

        B = taylor.sum(axis=0) @ B0

        B = jnp.vstack([jnp.zeros_like(B), B])
        B = B[swapdim, :]

        # B = dt * jnp.array([[0.], [0.], [0.], [10.]])

        # observation model
        C = jnp.array([[1., 0., 0., 0.], [0., 1., 0., 0.]])
        # C = jnp.array([[0., 0., 0., 0.] * delay + [1., -1., 0., 0.]])

        # noise model
        V = jnp.diag(jnp.array([1e-7, 0., sigma_w, 0.])) + motor_noise * B @ B.T
        W = jnp.diag(jnp.array([sigma, 0.]))

        # cost function
        Q = linalg.block_diag(jnp.array([[1., -1., 0., 0.], [-1., 1., 0., 0.], [0., 0., 0., 0.], [0., 0., 0., 0.]]))

        R = jnp.eye(B.shape[1]) * c

        dyn = Dynamics(A=A, B=B, C=C, V=V)
        act = Actor(A=A, B=B, C=C, V=V, W=W, Q=Q, R=R)

        super().__init__(actor=act, dynamics=dyn)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    model = DampedSpringTrackingFilter(sigma=40.)

    x = model.simulate(n=20, T=1000)

    plt.plot(x[:, 0, 0])
    plt.plot(x[:, 0, 1])
    plt.show()



import numpy as np
from scipy.signal import fftconvolve
from scipy.optimize import curve_fit


def xcorr(x, y, maxlags=60, normed=True):
    """ Compute cross correlation between two arrays along last axis of an array,
    treating other axes as batch dimensions

    Args:
        x, y: arrays of same dimensions
        maxlags: cutoff
        normed: normalize the correlations

    Returns:
        lags, correlations
    """
    Nx = x.shape[-1]

    correls = fftconvolve(x, y[..., ::-1], mode="full", axes=-1)

    if normed:
        correls = correls / np.sqrt(np.sum(x * x, axis=-1) * np.sum(y * y, axis=-1))[..., None]

    if maxlags is None:
        maxlags = Nx - 1

    if maxlags >= Nx or maxlags < 1:
        raise ValueError('maxlags must be None or strictly '
                         'positive < %d' % Nx)

    lags = np.arange(-maxlags, maxlags + 1)

    correls = correls[..., Nx - 1 - maxlags:Nx + maxlags]
    return lags, correls


import numpy as np


def dog(x, a1, a2, mu1, mu2, sigma1, sigma2):
    g = a1 / (sigma1 * np.sqrt(2 * np.pi)) * np.exp(-.5 * (x - mu1) ** 2 / sigma1 ** 2)
    h = a2 / (sigma2 * np.sqrt(2 * np.pi)) * np.exp(-.5 * (x - mu2) ** 2 / sigma2 ** 2)

    return g - h


def skewed_gabor(x, a, mu, sigma1, sigma2, w):
    # f = np.zeros_like(x)
    f = (x >= mu) * a * np.exp(-.5 * (x - mu) ** 2 / sigma1 ** 2) * np.sin(2 * np.pi * w * (x - mu)) + (
            x < mu) * a * np.exp(-.5 * (x - mu) ** 2 / sigma2 ** 2) * np.sin(
        2 * np.pi * w * (x - mu))

    return f


def fit_dog(x, y):
    res = curve_fit(dog, x, y)
    params = res[0]
    return dict(a1=params[0], a2=params[1], mu1=params[2], mu2=params[3], sigma1=params[4], sigma2=params[5])


def fit_skewed_gabor(x, y):
    res = curve_fit(skewed_gabor, x, y, max_nfev=5000, p0=np.array([0.5, 1., 5., 2., 1.]), method="trf",
                    bounds=(np.array([0., 0., .1, .1, .1]), np.array([1., 50., 50., 50., 5.])))
    params = res[0]
    return dict(a=params[0], mu=params[1], sigma1=params[2], sigma2=params[3], w=params[4])


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    from lqg.io.leap import load_tracking_data

    # x = np.linspace(0, 60)

    data, _ = load_tracking_data("S1", exp=2, delay=12, clip=60, subtract_mean=False, normalize=False,
                                 path="../../data/annotated_data")

    for data_i in data:
        x = data_i.transpose((1, 0, 2))

        lags, correls = xcorr(np.diff(x[..., 1], axis=0).T, np.diff(x[..., 0], axis=0).T, maxlags=60)

        plt.plot(lags, correls.mean(axis=0))

        res = fit_skewed_gabor(lags[60:], correls.mean(axis=0)[60:])

        plt.plot(lags, skewed_gabor(lags, **res))
        plt.show()



import numpyro.distributions as dist
import jax.numpy as jnp
from jax.lax import scan
import numpy as np

from lqg.riccati import discrete_riccati


def kalman_gain(A, C, V, W, T):
    P = discrete_riccati(A.T, C.T, V, W, T)

    S = C @ P @ C.T + W
    K = P @ C.T @ jnp.linalg.inv(S)

    return K


class KalmanFilter:
    def __init__(self, A, C, V, W):
        self.A = A
        self.C = C
        self.V = V
        self.W = W

    @property
    def xdim(self):
        return self.A.shape[0] * 2

    def K(self, T):
        K = kalman_gain(self.A, self.C, self.V @ self.V.T, self.W @ self.W.T, T=T)
        return K

    def simulate(self, n=1, T=100, seed=0):
        np.random.seed(seed)

        V = self.V
        W = self.W
        K = self.K(T)

        xs = []
        xhats = []

        xi = jnp.zeros((n, self.A.shape[0]))
        xhat = jnp.zeros((n, self.A.shape[0]))

        for t in range(T):
            # key, subkey = random.split(key)
            # xi = dist.MultivariateNormal(xi @ self.A.T, scale_tril=V).sample(subkey)
            xi = xi @ self.A.T + np.random.normal(size=(n, self.A.shape[0])) @ V.T

            # key, subkey = random.split(key)
            # yi = dist.MultivariateNormal(xi @ self.C.T, scale_tril=W).sample(subkey)
            yi = xi @ self.C.T + np.random.normal(size=(n, self.C.shape[0])) @ W.T

            xhat = xhat @ self.A.T + (yi - xhat @ self.A.T @ self.C.T) @ K.T

            xs.append(xi)
            xhats.append(xhat)

        x = jnp.stack(xs)
        xhat = jnp.stack(xhats)

        x = jnp.concatenate([x, xhat], axis=2)

        return jnp.concatenate([x[..., ::2], x[..., 1::2]], axis=2)

    def simulate_given_x(self, x, seed=0, x0=None, xhat0=None):
        T, n = x.shape

        np.random.seed(seed)

        A = np.array(self.A)

        V = np.array(self.V)
        W = np.array(self.W)
        K = np.array(self.K(T))

        xs = []
        xhats = []

        xi = np.zeros((n, A.shape[0])) if x0 is None else x0
        xhat = np.zeros((n, A.shape[0])) if xhat0 is None else xhat0

        xs.append(xi)
        xhats.append(xhat)

        for t in range(1, T):
            # key, subkey = random.split(key)
            # xi = dist.MultivariateNormal(xi @ self.A.T, scale_tril=V).sample(subkey)
            xi = xi @ A.T + np.random.normal(size=(n, A.shape[0])) @ V.T
            xi[:, 0] = x[t]

            # key, subkey = random.split(key)
            # yi = dist.MultivariateNormal(xi @ self.C.T, scale_tril=W).sample(subkey)
            yi = xi @ self.C.T + np.random.normal(size=(n, self.C.shape[0])) @ W.T

            xhat = xhat @ self.A.T + (yi - xhat @ self.A.T @ self.C.T) @ K.T

            xs.append(xi)
            xhats.append(xhat)

        x = jnp.stack(xs)
        xhat = jnp.stack(xhats)

        x = jnp.concatenate([x, xhat], axis=2)

        return jnp.concatenate([x[..., ::2], x[..., 1::2]], axis=2)

    def conditional_distribution(self, x):
        T, n, d = x.shape
        idx = list(range(d))
        swapdim = idx[::2] + idx[1::2]

        V = self.V
        W = self.W

        K = self.K(T)

        F = jnp.vstack(
            [jnp.hstack([self.A, jnp.zeros_like(self.A)]),
             jnp.hstack([K @ self.C @ self.A, self.A - K @ self.C @ self.A])])

        F = F[:, swapdim]
        F = F[swapdim, :]

        # TODO: this throws errors for the model with velocities..
        G = jnp.vstack([jnp.hstack([V, jnp.zeros_like(V)]), jnp.hstack([K @ self.C @ V, K @ W])])

        G = G[:, swapdim]
        G = G[swapdim, :]

        mu = jnp.zeros((n, d))
        Sigma = G @ G.T

        def f(carry, xt):
            mu, Sigma = carry

            mu = mu @ F.T + (xt - mu) @ jnp.linalg.inv(Sigma).T @ (F @ Sigma).T

            Sigma = F @ Sigma @ F.T + G @ G.T - (F @ Sigma) @ jnp.linalg.inv(Sigma) @ (Sigma @ F.T)
            return (mu, Sigma), (mu, Sigma)

        _, (mu, Sigma) = scan(f, (mu, Sigma), x)

        return dist.MultivariateNormal(mu.transpose((1, 0, 2)), Sigma)

    def log_likelihood(self, x):
        return self.conditional_distribution(x[:-1]).log_prob(x[1:].transpose((1, 0, 2)))

    def mean_given_x(self, x):
        T, n, d = x.shape

        A = np.array(self.A)
        C = np.array(self.C)

        K = np.array(self.K(T))

        xs = []
        xhats = []

        target = x[:, :, ::2]

        xi = target[0]  # np.zeros((n, self.dynamics.A.shape[0])) if x0 is None else x0
        xi = np.nan_to_num(xi)

        # TODO: this is a problem for models in which the subjective dimensionality is different
        xhat = x[0, :, 1::2]  # np.zeros((n, self.actor.A.shape[0])) if xhat0 is None else xhat0
        xhat = np.nan_to_num(xhat)

        xs.append(xi)
        xhats.append(xhat)

        for t in range(1, T):
            # xi = xi @ A.T
            xi = target[t]

            yi = xi @ C.T

            xhat = xhat @ A.T + (yi - xhat @ A.T @ C.T) @ K.T

            xs.append(xi)
            xhats.append(xhat)

        x = np.stack(xs)
        xhat = np.stack(xhats)

        x = np.concatenate([x, xhat], axis=2)

        return np.concatenate([x[..., ::2], x[..., 1::2]], axis=2)



import jax.numpy as jnp
from jax import random, lax
import numpyro
from numpyro.infer import SVI, Trace_ELBO

from lqg.tracking import OneDimModel
from lqg.infer.models import lqg_model, common_lqg_model


def guide(x, model_type, process_noise=None, dt=None, **fixed_params):
    pass


def max_likelihood(x, model=OneDimModel, numpyro_fn=lqg_model, process_noise=1., dt=1. / 60, steps=2_000,
                   step_size=0.01, **fixed):
    # optim = numpyro.optim.ClippedAdam(step_size=step_size)
    optim = numpyro.optim.Adam(step_size=step_size)

    # optim = numpyro.optim.Minimize()
    svi = SVI(numpyro_fn, guide, optim, Trace_ELBO(), x=x, model_type=model, process_noise=process_noise, dt=dt,
              **fixed)
    params, state, losses = svi.run(random.PRNGKey(3), steps)

    return params, losses


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import arviz as az

    c_true = .5
    motor_noise_true = .25
    sigma_true = 8.
    prop_noise_true = 2.
    lqg = OneDimModel(process_noise=1.,
                      sigma=sigma_true, motor_noise=motor_noise_true, c=c_true, prop_noise=prop_noise_true)

    results = dict(c=[], sigma=[], motor_noise=[], prop_noise=[])
    for seed in range(1):
        x = lqg.simulate(n=20, T=500, seed=seed)

        params, losses = max_likelihood(x, steps=1)

        for key, item in params.items():
            results[key].append(item)

    for key in results.keys():
        results[key] = jnp.stack(results[key])

    az.plot_pair(results, figsize=(8, 8))
    plt.show()



import jax.numpy as jnp
# from fax import implicit
from jax.lax import scan


# def solve_dare(A, B, Q, R):
#     def _make_riccati_operator(params):
#         A, B, Q, R = params
#
#         def _riccati_operator(P):
#             X = R + B.T @ P.T @ B
#             Y = B.T @ P @ A
#             return (A.T @ P @ A) - ((A.T @ P @ B) @ jnp.linalg.solve(X, Y)) + Q
#
#         return _riccati_operator
#
#     solution = implicit.two_phase_solve(_make_riccati_operator, jnp.eye(A.shape[0]), (A, B, Q, R))
#     return solution


def discrete_riccati(A, B, Q, R, T):
    def riccati_iter(S, t):
        S = A.T @ (S - S @ B @ jnp.linalg.inv(B.T @ S @ B + R) @ B.T @ S) @ A + Q
        return S, S

    S, _ = scan(riccati_iter, Q, jnp.arange(T))

    # alternative method, does not seem to be faster though
    # S = solve_dare(A, B, Q, R)

    return S


import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist

from lqg.infer.models import common_lqg_model

prior = {"c": dist.Gamma(0.01, 0.01), "motor_noise": dist.Gamma(0.01, 0.01),
         "sigma_0": dist.Gamma(0.01, 0.01), "sigma_1": dist.Gamma(0.01, 0.01), "sigma_2": dist.Gamma(0.01, 0.01),
         "sigma_3": dist.Gamma(0.01, 0.01), "sigma_4": dist.Gamma(0.01, 0.01), "sigma_5": dist.Gamma(0.01, 0.01)
         }
hierarchical_lqg_model = numpyro.handlers.lift(common_lqg_model, prior)


def lqg_guide(x):
    # control costs
    delta_c = numpyro.param("c_loc", jnp.array(1.), constraint=dist.constraints.positive)
    c = numpyro.sample("c", dist.Delta(delta_c))

    # observation noise
    delta_sigma = numpyro.param("sigma_loc", jnp.array(1.), constraint=dist.constraints.positive)
    sigma = numpyro.sample("sigma", dist.Delta(delta_sigma))

    # motor variability
    delta_motor_noise = numpyro.param("motor_noise_loc", jnp.array(1.), constraint=dist.constraints.positive)
    motor_noise = numpyro.sample("motor_noise", dist.Delta(delta_motor_noise))


def hierarchical_guide(x):
    Nc, N, T, d = x.shape

    # control costs
    delta_c = numpyro.param("c_loc", jnp.array(1.), constraint=dist.constraints.positive)
    c = numpyro.sample("c", dist.Delta(delta_c))

    # motor variability
    delta_motor_noise = numpyro.param("motor_noise_loc", jnp.array(1.), constraint=dist.constraints.positive)
    motor_noise = numpyro.sample("motor_noise", dist.Delta(delta_motor_noise))

    for n in range(Nc):
        # observation noise
        delta_sigma_n = numpyro.param(f"sigma_{n}_loc", jnp.array(1.), constraint=dist.constraints.positive)
        sigma_n = numpyro.sample(f"sigma_{n}", dist.Delta(delta_sigma_n))



import numpy as np
import numpyro.handlers
from jax import numpy as jnp
from jax.scipy.stats import norm
from numpyro import distributions as dist


def prior():
    return {"c": dist.HalfNormal(2.),
            "sigma": dist.HalfNormal(50.),
            "motor_noise": dist.HalfNormal(1.),
            "prop_noise": dist.HalfNormal(12.5),
            "vel_noise": dist.HalfCauchy(10.),
            "subj_noise": dist.HalfNormal(1.),
            "subj_vel_noise": dist.HalfNormal(2.),
            "subj_k": dist.Gamma(.01, .01),
            "subj_lmbd": dist.Gamma(.01, .01),
            "sigma_v": dist.HalfCauchy(50.),
            "sigma_h": dist.HalfCauchy(50.),
            "sigma_z": dist.HalfCauchy(50.),
            "motor_noise_v": dist.HalfCauchy(1.),
            "motor_noise_h": dist.HalfCauchy(1.),
            "prop_noise_v": dist.HalfCauchy(1.),
            "prop_noise_h": dist.HalfCauchy(1.),
            "c_v": dist.HalfCauchy(50.),
            "c_h": dist.HalfCauchy(50.),
            "sigma_0": dist.HalfNormal(50.),
            "sigma_1": dist.HalfNormal(50.),
            "sigma_2": dist.HalfNormal(50.),
            "sigma_3": dist.HalfNormal(50.),
            "sigma_4": dist.HalfNormal(50.),
            "sigma_5": dist.HalfNormal(50.),
            "sigma_test": dist.Gamma(0.01, 0.01)
            }


def lognormal_params(mu, sigma):
    return np.log(mu ** 2 / np.sqrt(mu ** 2 + sigma ** 2)), np.log(1 + sigma ** 2 / mu ** 2)


def lognormal_from_quantiles(x1, x2, p1=0.05, p2=0.95):
    """ Compute the parameters of a log-normal distribution, such that F(x1) = p1 and F(x2) = p2

    Args:
        x1: lower value
        x2: upper value
        p1: lower probability
        p2: upper probability

    Returns:
        mu, sigma (parameters of the log-normal distribution
    """
    sigma = (jnp.log(x2) - jnp.log(x1)) / (norm.ppf(p2) - norm.ppf(p1))
    mu = (jnp.log(x2) * norm.ppf(p2) - jnp.log(x1) * norm.ppf(p1)) / (norm.ppf(p2) - norm.ppf(p1))
    return mu, sigma


def sample_params(prior_dict):
    params = {}
    for param, d in prior_dict.items():
        params[param] = numpyro.sample(param, d)

    return params


if __name__ == '__main__':
    import numpyro
    from jax import random
    import arviz as az
    import matplotlib.pyplot as plt
    import pandas as pd

    from lqg.tracking import OneDimModel as LQG

    # test priors
    param_names = ["sigma", "sigma_test"]

    x = jnp.linspace(0, 100)

    f, axes = plt.subplots(1, len(param_names), sharey=True)
    prior_dict = prior()
    for ax, name in zip(axes, param_names):
        p = prior_dict[name]

        ax.plot(x, jnp.exp(p.log_prob(x)))
        ax.set_title(name)

    plt.show()

    # # prior predictive checks
    # plist = []
    # for seed in range(25):
    #     params = numpyro.handlers.seed(sample_params, rng_seed=seed)(prior_dict)
    #     params = {k: v.item() for k, v in params.items() if k in param_names}
    #     plist.append(params)
    #
    #     lqg = LQG(**params)
    #
    #     x = lqg.simulate(n=20, T=1000, seed=seed)
    #
    #     #     # visualize trajectories
    #     #     plt.plot(x[:, 0, 0])
    #     #     plt.plot(x[:, 0, 1])
    #     #     plt.xlabel("time")
    #     #     plt.ylabel("position")
    #     #     plt.show()
    #
    #     lags, correls = xcorr(jnp.diff(x[..., 1], axis=0).T, jnp.diff(x[..., 0], axis=0).T, maxlags=60)
    #     plt.plot(lags[50:], correls.mean(axis=0)[50:])
    #
    # plt.show()
    #
    # # df = pd.DataFrame(plist)
    #
    # # import seaborn as sns
    # #
    # # sns.pairplot(df)
    # # plt.show()
    #
    # # check HDIs of priors.
    # print(az.hdi(np.array(numpyro.distributions.HalfCauchy(50.).sample(random.PRNGKey(0), (10_000,)))))



from lqg.tracking.basic import OneDimModel, TwoDimModel, VelocityModel, VelocityDiffModel, DimModel, DiffModel, \
    NoiseFreeModel, CostlessModel
from lqg.tracking.eye import DampedSpringModel, DampedSpringVelocityModel, DampedSpringTwoDimModel, \
    DampedSpringDiffModel, DampedSpringSubjectiveVelocityModel, DampedSpringSubjectiveModel, \
    DampedSpringTwoDimSubjectiveModel, DampedSpringTwoDimFullModel, DampedSpringTrackingFilter, \
    DampedSpringCostlessModel, DampedSpringTwoDimCostlessModel
from lqg.tracking.leap import Independent3DModel, Independent3DVelocityModel
from lqg.tracking.kf import TrackingFilter, TwoDimTrackingFilter
from lqg.tracking.subjective import SubjectiveModel, SubjectiveVelocityModel, SubjectiveVelocityDiffModel, \
    DelayedSubjectiveVelocityModel



import numpy as np
import jax.numpy as jnp
import numpyro.distributions as dist
from jax.lax import scan

from lqg.kalman import kalman_gain
from lqg.lqr import control_law


class Dynamics:
    def __init__(self, A, B, C, V):
        self.A = A
        self.B = B
        self.C = C

        self.V = V

    def simulate(self, n=1, T=100, seed=0):
        np.random.seed(seed)
        V = self.V

        xs = []

        xi = jnp.zeros((n, self.A.shape[0]))

        for t in range(0, T):
            xi = xi @ self.A.T + np.random.normal(size=(n, self.A.shape[0])) @ V.T

            xs.append(xi)

        return jnp.stack(xs)


class Actor:
    def __init__(self, A, B, C, V, W, Q, R):
        self.A = A
        self.B = B
        self.C = C

        self.V = V
        self.W = W

        self.Q = Q
        self.R = R

    def L(self, T=100):
        return control_law(self.A, self.B, self.Q, self.R, T=T)

    def K(self, T=100):
        K = kalman_gain(self.A, self.C, self.V @ self.V.T, self.W @ self.W.T, T=T)
        return K


class System:
    def __init__(self, actor, dynamics):
        self.actor = actor
        self.dynamics = dynamics

    @property
    def xdim(self):
        return self.dynamics.A.shape[0]

    @property
    def udim(self):
        return self.dynamics.B.shape[1]

    def simulate(self, n=1, T=100, seed=0, x0=None, xhat0=None, return_all=False):
        np.random.seed(seed)

        A = np.array(self.dynamics.A)
        B = np.array(self.dynamics.B)
        C = np.array(self.dynamics.C)

        L = np.array(self.actor.L(T=T))
        K = np.array(self.actor.K(T=T))

        W = np.array(self.actor.W)
        V = np.array(self.dynamics.V)

        A_act = np.array(self.actor.A)
        B_act = np.array(self.actor.B)
        C_act = np.array(self.actor.C)

        xi = np.zeros((n, self.dynamics.A.shape[0])) if x0 is None else x0
        xhat = np.zeros((n, self.actor.A.shape[0])) if xhat0 is None else xhat0
        # xhat = np.zeros((n, self.actor.A.shape[0]))
        # xi = np.zeros((n, self.dynamics.A.shape[0]))
        # xi = np.array(np.tile(self.x0, (n, 1)))
        # xhat = np.array(np.tile(self.xhat0, (n, 1)))
        u = np.zeros((n, self.actor.B.shape[1])) if xhat0 is None else (-xhat0 @ L.T)

        xs = []

        if return_all:
            xhats = []
            ys = [] # [xi @ C.T + np.random.normal(size=(n, C.shape[0])) @ W.T]
            us = []

        for t in range(T):


            xi = xi @ A.T + u @ B.T + np.random.normal(size=(n, A.shape[0])) @ V.T
            # xi = random.multivariate_normal(subkey, xi @ self.A.T + u @ self.B.T, V)

            yi = xi @ C.T + np.random.normal(size=(n, C.shape[0])) @ W.T
            # yi = random.multivariate_normal(subkey, xi @ self.C.T, W)

            x_pred = xhat @ A_act.T + u @ B_act.T

            xhat = x_pred + (yi - x_pred @ C_act.T) @ K.T

            u = - xhat @ L.T

            xs.append(xi)

            if return_all:
                xhats.append(xhat)
                ys.append(yi)
                us.append(u)

        if return_all:
            return np.stack(xs), np.stack(xhats), np.stack(ys), np.stack(us)
        else:
            return np.stack(xs)

    def simulate_given_x(self, x, seed=0, x0=None, xhat0=None, return_all=False):

        T, n = x.shape

        np.random.seed(seed)

        A = np.array(self.dynamics.A)
        B = np.array(self.dynamics.B)
        C = np.array(self.dynamics.C)

        L = np.array(self.actor.L(T=T))
        K = np.array(self.actor.K(T=T))

        W = np.array(self.actor.W)
        V = np.array(self.dynamics.V)

        A_act = np.array(self.actor.A)
        B_act = np.array(self.actor.B)
        C_act = np.array(self.actor.C)

        xs = []

        xi = np.zeros((n, self.dynamics.A.shape[0])) if x0 is None else x0
        xhat = np.zeros((n, self.actor.A.shape[0])) if xhat0 is None else xhat0
        # xhat = np.zeros((n, self.actor.A.shape[0]))
        # xi = np.zeros((n, self.dynamics.A.shape[0]))
        # xi = np.array(np.tile(self.x0, (n, 1)))
        # xhat = np.array(np.tile(self.xhat0, (n, 1)))

        xs.append(xi)

        if return_all:
            xhats = [xhat]

        for t in range(1, T):
            u = - xhat @ L.T

            xi = xi @ A.T + u @ B.T + np.random.normal(size=(n, A.shape[0])) @ V.T
            xi[:, 0] = x[t]
            # xi = random.multivariate_normal(subkey, xi @ self.A.T + u @ self.B.T, V)

            yi = xi @ C.T + np.random.normal(size=(n, C.shape[0])) @ W.T
            # yi = random.multivariate_normal(subkey, xi @ self.C.T, W)

            x_pred = xhat @ A_act.T + u @ B_act.T

            xhat = x_pred + (yi - x_pred @ C_act.T) @ K.T

            xs.append(xi)
            if return_all:
                xhats.append(xhat)

        if return_all:
            return jnp.stack(xs), jnp.stack(xhats)
        else:
            return jnp.stack(xs)

    def conditional_moments(self, x, mu0=None):
        """ Conditional distribution p(x | theta)

                Args:
                    self: LQG
                    x: time series of shape T (time steps), n (trials), d (dimensionality)

                Returns:
                    numpyro.distributions.MultivariateNormal
                """
        T, n, d = x.shape

        # xdim = self.dynamics.A.shape[0]
        # xhatdim = self.actor.A.shape[0]

        L = self.actor.L(T)

        K = self.actor.K(T)

        F = jnp.vstack(
            [jnp.hstack([self.dynamics.A,
                         -self.dynamics.B @ L]),
             jnp.hstack([K @ self.dynamics.C @ self.dynamics.A,
                         self.actor.A - self.actor.B @ L - K @ self.actor.C @ self.actor.A])])

        # G = jnp.vstack([jnp.hstack([V, jnp.zeros((d, d // 2)), jnp.zeros((d, d))]),
        #                 jnp.hstack([K @ self.C @ V, K @ W, self.E - K @ self.C @ self.E])])

        G = jnp.vstack([jnp.hstack([self.dynamics.V, jnp.zeros_like(self.dynamics.C.T)]),
                        jnp.hstack([K @ self.dynamics.C @ self.dynamics.V, K @ self.actor.W])])

        # mu = jnp.hstack((x[0], x[0]))
        mu = jnp.zeros((n, self.dynamics.A.shape[0] + self.actor.A.shape[0])) if mu0 is None else mu0
        Sigma = G @ G.T

        # Sigma += jnp.eye(Sigma.shape[0]) * 1e-7

        # return (x[0] - mu[:, :d]) # @ jnp.linalg.inv(Sigma[:d, :d]).T @ (F @ Sigma)[:, :d].T

        def f(carry, xt):
            mu, Sigma = carry

            mu = mu @ F.T + (xt - mu[:, :d]) @ jnp.linalg.inv(Sigma[:d, :d]).T @ (F @ Sigma)[:, :d].T

            Sigma = F @ Sigma @ F.T + G @ G.T - (F @ Sigma)[:, :d] @ jnp.linalg.inv(Sigma[:d, :d]) @ (Sigma @ F.T)[:d,
                                                                                                     :]
            return (mu, Sigma), (mu, Sigma)

        _, (mu, Sigma) = scan(f, (mu, Sigma), x)
        return mu, Sigma

    def conditional_distribution(self, x):
        T, n, d = x.shape
        mu, Sigma = self.conditional_moments(x)
        return dist.MultivariateNormal(mu[:, :, :d].transpose((1, 0, 2)), Sigma[:, :d, :d])

    def log_likelihood(self, x):
        return self.conditional_distribution(x[:-1]).log_prob(x[1:].transpose((1, 0, 2)))

    def mean_given_x(self, x):
        T, n, d = x.shape

        A = np.array(self.dynamics.A)
        B = np.array(self.dynamics.B)
        C = np.array(self.dynamics.C)

        L = np.array(self.actor.L(T=T))
        K = np.array(self.actor.K(T=T))

        A_act = np.array(self.actor.A)
        B_act = np.array(self.actor.B)
        C_act = np.array(self.actor.C)

        xs = []

        xi = x[0]  # np.zeros((n, self.dynamics.A.shape[0])) if x0 is None else x0
        xi = np.nan_to_num(xi)

        # TODO: this is a problem for models in which the subjective dimensionality is different
        xhat = x[0]  # np.zeros((n, self.actor.A.shape[0])) if xhat0 is None else xhat0
        xhat = np.nan_to_num(xhat)

        xs.append(xi)

        for t in range(1, T):
            u = - xhat @ L.T

            xi = xi @ A.T + u @ B.T
            xi[~np.isnan(x[t])] = x[t][~np.isnan(x[t])]
            # xi[:, 0:-2:2] = x[t]

            x_pred = xhat @ A_act.T + u @ B_act.T

            xhat = x_pred + (xi @ C.T - x_pred @ C_act.T) @ K.T

            xs.append(xi)

        return np.stack(xs)


class LQG(System):
    def __init__(self, A, B, C, V, W, Q, R):
        dynamics = Dynamics(A, B, C, V)
        actor = Actor(A, B, C, V, W, Q, R)
        super().__init__(actor=actor, dynamics=dynamics)



from jax import numpy as jnp

from lqg.kalman import KalmanFilter


class TrackingFilter(KalmanFilter):
    def __init__(self, dim=1, process_noise=1., sigma=6., dt=1. / 60.):
        A = jnp.eye(dim)
        C = jnp.eye(dim)
        V = jnp.eye(dim) * process_noise
        W = jnp.eye(dim) * sigma
        super().__init__(A, C, V, W)


class TwoDimTrackingFilter(KalmanFilter):
    def __init__(self, process_noise=1., sigma_v=6., sigma_h=6., dt=1. / 60.):
        A = jnp.eye(2)
        C = jnp.eye(2)
        V = jnp.eye(2) * process_noise
        W = jnp.diag(jnp.array([sigma_h, sigma_v]))
        super().__init__(A, C, V, W)



"""
A collection of helper functions for optimization with JAX.
"""
import numpy as onp
import scipy.optimize
from jax import grad, jit
from jax.flatten_util import ravel_pytree

from jax.config import config

config.update('jax_enable_x64', True)


def minimize(fun, x0,
             method=None,
             args=(),
             bounds=None,
             constraints=(),
             tol=None,
             callback=None,
             options=None):
    """
    A simple wrapper for scipy.optimize.minimize using JAX.

    Args:
        fun: The objective function to be minimized, written in JAX code
        so that it is automatically differentiable.  It is of type,
            ```fun: x, *args -> float```
        where `x` is a PyTree and args is a tuple of the fixed parameters needed
        to completely specify the function.

        x0: Initial guess represented as a JAX PyTree.

        args: tuple, optional. Extra arguments passed to the objective function
        and its derivative.  Must consist of valid JAX types; e.g. the leaves
        of the PyTree must be floats.

        _The remainder of the keyword arguments are inherited from
        `scipy.optimize.minimize`, and their descriptions are copied here for
        convenience._

        method : str or callable, optional
        Type of solver.  Should be one of
            - 'Nelder-Mead' :ref:`(see here) <optimize.minimize-neldermead>`
            - 'Powell'      :ref:`(see here) <optimize.minimize-powell>`
            - 'CG'          :ref:`(see here) <optimize.minimize-cg>`
            - 'BFGS'        :ref:`(see here) <optimize.minimize-bfgs>`
            - 'Newton-CG'   :ref:`(see here) <optimize.minimize-newtoncg>`
            - 'L-BFGS-B'    :ref:`(see here) <optimize.minimize-lbfgsb>`
            - 'TNC'         :ref:`(see here) <optimize.minimize-tnc>`
            - 'COBYLA'      :ref:`(see here) <optimize.minimize-cobyla>`
            - 'SLSQP'       :ref:`(see here) <optimize.minimize-slsqp>`
            - 'trust-constr':ref:`(see here) <optimize.minimize-trustconstr>`
            - 'dogleg'      :ref:`(see here) <optimize.minimize-dogleg>`
            - 'trust-ncg'   :ref:`(see here) <optimize.minimize-trustncg>`
            - 'trust-exact' :ref:`(see here) <optimize.minimize-trustexact>`
            - 'trust-krylov' :ref:`(see here) <optimize.minimize-trustkrylov>`
            - custom - a callable object (added in version 0.14.0),
              see below for description.
        If not given, chosen to be one of ``BFGS``, ``L-BFGS-B``, ``SLSQP``,
        depending if the problem has constraints or bounds.

        bounds : sequence or `Bounds`, optional
            Bounds on variables for L-BFGS-B, TNC, SLSQP, Powell, and
            trust-constr methods. There are two ways to specify the bounds:
                1. Instance of `Bounds` class.
                2. Sequence of ``(min, max)`` pairs for each element in `x`. None
                is used to specify no bound.
            Note that in order to use `bounds` you will need to manually flatten
            them in the same order as your inputs `x0`.

        constraints : {Constraint, dict} or List of {Constraint, dict}, optional
            Constraints definition (only for COBYLA, SLSQP and trust-constr).
            Constraints for 'trust-constr' are defined as a single object or a
            list of objects specifying constraints to the optimization problem.
            Available constraints are:
                - `LinearConstraint`
                - `NonlinearConstraint`
            Constraints for COBYLA, SLSQP are defined as a list of dictionaries.
            Each dictionary with fields:
                type : str
                    Constraint type: 'eq' for equality, 'ineq' for inequality.
                fun : callable
                    The function defining the constraint.
                jac : callable, optional
                    The Jacobian of `fun` (only for SLSQP).
                args : sequence, optional
                    Extra arguments to be passed to the function and Jacobian.
            Equality constraint means that the constraint function result is to
            be zero whereas inequality means that it is to be non-negative.
            Note that COBYLA only supports inequality constraints.

            Note that in order to use `constraints` you will need to manually flatten
            them in the same order as your inputs `x0`.

        tol : float, optional
            Tolerance for termination. For detailed control, use solver-specific
            options.

        options : dict, optional
            A dictionary of solver options. All methods accept the following
            generic options:
                maxiter : int
                    Maximum number of iterations to perform. Depending on the
                    method each iteration may use several function evaluations.
                disp : bool
                    Set to True to print convergence messages.
            For method-specific options, see :func:`show_options()`.

        callback : callable, optional
            Called after each iteration. For 'trust-constr' it is a callable with
            the signature:
                ``callback(xk, OptimizeResult state) -> bool``
            where ``xk`` is the current parameter vector represented as a PyTree,
             and ``state`` is an `OptimizeResult` object, with the same fields
            as the ones from the return. If callback returns True the algorithm
            execution is terminated.

            For all the other methods, the signature is:
                ```callback(xk)```
            where `xk` is the current parameter vector, represented as a PyTree.

    Returns:
        res : The optimization result represented as a ``OptimizeResult`` object.
        Important attributes are:
            ``x``: the solution array, represented as a JAX PyTree
            ``success``: a Boolean flag indicating if the optimizer exited successfully
            ``message``: describes the cause of the termination.
        See `scipy.optimize.OptimizeResult` for a description of other attributes.

    """
    # Use tree flatten and unflatten to convert params x0 from PyTrees to flat arrays
    x0_flat, unravel = ravel_pytree(x0)

    # Wrap the objective function to consume flat _original_
    # numpy arrays and produce scalar outputs.
    def fun_wrapper(x_flat, *args):
        x = unravel(x_flat)
        return float(fun(x, *args))

    # Wrap the gradient in a similar manner
    jac = jit(grad(fun))

    def jac_wrapper(x_flat, *args):
        x = unravel(x_flat)
        g_flat, _ = ravel_pytree(jac(x, *args))
        return onp.array(g_flat)

    # Wrap the callback to consume a pytree
    def callback_wrapper(x_flat, *args):
        if callback is not None:
            x = unravel(x_flat)
            return callback(x, *args)

    # Minimize with scipy
    results = scipy.optimize.minimize(fun_wrapper,
                                      x0_flat,
                                      args=args,
                                      method=method,
                                      jac=jac_wrapper,
                                      callback=callback_wrapper,
                                      bounds=bounds,
                                      constraints=constraints,
                                      tol=tol,
                                      options=options)

    # pack the output back into a PyTree
    results["x"] = unravel(results["x"])
    return results



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lqg",
    version="0.0.1",
    author="Dominik Straub",
    author_email="dominik.straub@tu-darmstadt.de",
    description="(Inverse) optimal control for linear quadratic Gaussian systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dominikstrb/lqg",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        "numpy>=1.20.3",
        "scipy>=1.6.3",
        "matplotlib>=3.4.2",
        "numpyro==0.8.0",
        "jax==0.2.25",
        "jaxlib==0.1.74",
        "arviz>=0.11.2",
    ],
)



from jax import numpy as jnp

from lqg.riccati import discrete_riccati


def control_law(A, B, Q, R, T):
    S = discrete_riccati(A, B, Q, R, T)

    L = jnp.linalg.inv(B.T @ S @ B + R) @ B.T @ S @ A

    return L